package algs31;

import algs13.Queue;

// NB: Implementation of methods copied from algs32.BST
public class BinarySearchTreeST<K extends Comparable<? super K>, V> implements algs31.SymbolTable<K, V> {

    private static class Node<K extends Comparable<? super K>,V> {
        public final K key;       // sorted by key
        public V val;             // associated data
        public Node<K,V> left, right;  // left and right subtrees
        public int N;             // number of nodes in subtree

        public Node(K key, V val, int N) {
            this.key = key;
            this.val = val;
            this.N = N;
        }
    }
    
	private Node<K,V> root;             // root of BST

    // Recursive!
    public void put(K key, V val) {
        if (val == null) { delete(key); return; }
        root = put(root, key, val);
    }

    private Node<K,V> put(Node<K,V> x, K key, V val) {
        if (x == null) return new Node<>(key, val, 1);
        int cmp = key.compareTo(x.key);
        if      (cmp < 0)
            x.left  = put(x.left,  key, val);
        else if (cmp > 0)
            x.right = put(x.right, key, val);
        else
            x.val   = val;
        x.N = 1 + size(x.left) + size(x.right);
        return x;
    }

    @Override
    public V get(K key) { return get(root, key); }
    private V get(Node<K,V> x, K key) {
        if (x == null) return null;
        int cmp = key.compareTo(x.key);
        if      (cmp < 0) return get(x.left, key);
        else if (cmp > 0) return get(x.right, key);
        else              return x.val;
    }
    
    @Override
    public void delete(K key) {
        root = delete(root, key);
    }

    private Node<K,V> delete(Node<K,V> x, K key) {
        if (x == null) return null;
        int cmp = key.compareTo(x.key);
        if      (cmp < 0) x.left  = delete(x.left,  key);
        else if (cmp > 0) x.right = delete(x.right, key);
        else {
            if (x.right == null) return x.left;
            if (x.left  == null) return x.right;
            Node<K,V> t = x;
            x = min(t.right);
            x.right = deleteMin(t.right);
            x.left = t.left;
        }
        x.N = size(x.left) + size(x.right) + 1;
        return x;
    }

    private Node<K,V> deleteMin(Node<K,V> x) {
        if (x.left == null) return x.right;
        x.left = deleteMin(x.left);
        x.N = size(x.left) + size(x.right) + 1;
        return x;
    }

    @Override
    public boolean isEmpty() {
        return size() == 0;
    }

    @Override
    // return number of key-value pairs in BST
    public int size() { return size(root); }

    // return number of key-value pairs in BST rooted at x
    private int size(Node<K,V> x) {
        if (x == null) return 0;
        else return x.N;
    }

    @Override
    public Iterable<K> keys() {
        Queue<K> q = new Queue<>();
        inOrder(root, q);
        return q;
    }

    private void inOrder(Node<K,V> x, Queue<K> q) {
        if (x == null) return;
        inOrder(x.left, q);
        inOrder(x.right, q);
        q.enqueue(x.key);
    }

    @Override
    public boolean contains(K key) {
        return get(key) != null;
    }

    public K min() {
        if (isEmpty()) return null;
        return min(root).key;
    }

    private Node<K,V> min(Node<K,V> x) {
        if (x.left == null) return x;
        else                return min(x.left);
    }    
}
